@@ -259,6 +259,8 @@ void SessionState::handleContent(AMQFrame& frame, const SequenceNumber& id)
             header.setEof(false);
             msg->getFrames().append(header);
         }
+        if (broker.isTimestamping())
+            msg->setTimestamp();
         msg->setPublisher(&getConnection());
         msg->getIngressCompletion().begin();
         semanticState.handle(msg);
@@ -346,10 +348,8 @@ void SessionState::completeRcvMsg(SequenceNumber id,
     }
 
     // if the sender has requested immediate notification of the completion...
-    if (requiresSync) {
+    if (requiresSync || callSendCompletion) {
         sendAcceptAndCompletion();
-    } else if (callSendCompletion) {
-        sendCompletion();
     }
 }
 
